/*******************************************************************************
 * exception.H - Exception base class
 *
 * $Id: exception.H,v 2.0 2001/05/23 11:49:09 mscheid Exp $
 *
 * Generally used exception classes and general functions and macros to handle
 * them.
 *
 * Author: Matthias Scheidegger <mscheid@iam.unibe.ch>
 *
 * Copyright 2001, Matthias Scheidegger
 *
 ******************************************************************************/

#ifndef __EXCEPTION_H
#define __EXCEPTION_H

#include <stdarg.h>
#include <errno.h>
extern int errno;


class Exception
{
    /* basic exception class, which should be subtyped using NEW_EXCEPTION
       (see below). Use the template do_throw functions to throw an
       exception conveniently. */

  public:
    Exception() : msg(0) {}
    Exception(const Exception &other);
    Exception& operator=(const Exception &other);
    virtual ~Exception();

    virtual void set_string(const char *fmt, va_list vl);
    virtual void set_string(int error_number, const char *func);
    virtual void set_string(int error_number);
    virtual void set_string();

    char *msg;
    
  protected:
    virtual const char* class_name() const { return "Exception"; }
    virtual const char* standard_text() const { return "unspecified"; }
    static char* prepend_string(const char *pre, const char *str);

    static const int msg_len = 256;
};



extern void warn(const char *fmt, ...);
/* print the message to stderr and return */


extern void die(const char *fmt, ...);
/* print the message to stderr and exit with return value 2
   (i.e. failed execution) */


extern void mem_handler();
/* error handler for operator new. Can also be called manually (e.g. when using
   C-lib functions */


template <class T>
void do_throw(char *fmt, ...)
{
    /* Create an Exception of type T, set its msg variable and
       throw it. */

    va_list args;
    T exc;
    Exception &base_exc = exc;
    
    va_start(args, fmt);
    base_exc.set_string(fmt, args);
    va_end(args);
    throw(exc);
}



template <class T>
void do_throw(int error_number)
{
    /* Create an Exception of type T, set its msg variable using the error string from
       the standard C library and throw it. */

    T exc;

    exc.set_string(error_number);
    throw(exc);
}



template <class T>
void do_throw(int error_number, const char *func)
{
    /* Create an Exception of type T, set its msg variable using the error string from
       the standard C library, prepend the function string and throw it. */

    T exc;

    exc.set_string(error_number, func);
    throw(exc);
}



template <class T>
void do_throw()
{
    /* Create an Exception of type T, set its msg variable using its standard_text function
       and throw it */

    T exc;

    exc.set_string();
    throw(exc);
}



/* Macro to create trivial subclasses to Exception (or other childs of it)
   Used inside modules that need their own set of exception classes */

#define NEW_EXCEPTION(CHILD, PARENT) \
class CHILD : public PARENT\
{\
  public:\
    CHILD() : PARENT() {}\
    CHILD(const CHILD &other) : PARENT(other) {}\
    CHILD& operator=(const CHILD &other) { return (CHILD&)PARENT::operator=(other); }\
\
  protected:\
    virtual const char* class_name() const { return #CHILD; }\
};



#define NEW_EXCEPTION_WITH_TEXT(CHILD, PARENT, TEXT) \
class CHILD : public PARENT\
{\
  public:\
    CHILD() : PARENT() {}\
    CHILD(const CHILD &other) : PARENT(other) {}\
    CHILD& operator=(const CHILD &other) { return (CHILD&)PARENT::operator=(other); }\
\
  protected:\
    virtual const char* class_name() const { return #CHILD; }\
    virtual const char* standard_text() const { return TEXT; }\
};



/** standard exception classes ************************************************/

NEW_EXCEPTION_WITH_TEXT(OutOfMemory, Exception, "Memory allocation failed")
NEW_EXCEPTION(Assertion, Exception)
NEW_EXCEPTION(FileException, Exception)

#endif
